from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from cmcrameri import cm
from matplotlib.ticker import MultipleLocator

from proteus.interior.wrapper import read_interior_data
from proteus.utils.plot import MyFuncFormatter, latex_float, sample_output

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)

def plot_interior(output_dir: str, times: list | np.ndarray, data:list, module:str, plot_format: str="pdf"):

    if np.amax(times) < 2:
        log.debug("Insufficient data to make plot_interior")
        return
    if module not in ["spider","aragog"]:
        log.warning(f"Cannot make interior plot for module '{module}'")
        return

    log.info("Plot interior")

    # Init figure
    scale = 1.0
    fig,axs = plt.subplots(1,4, figsize=(10*scale,6*scale), sharey=True)

    # Create colormapping stuff
    norm = mpl.colors.LogNorm(vmin=max(1,times[0]), vmax=times[-1])
    sm = plt.cm.ScalarMappable(cmap=cm.batlowK_r, norm=norm)
    sm.set_array([])

    visc_min, visc_max = 1e99, 0
    flux_min, flux_max = 1e99, 0

    lw=1.5

    # loop over times
    for i,time in enumerate(times):

        # Get decorators
        label = latex_float(time)+" yr"
        color = sm.to_rgba(time)

        # Get interior data for this time
        ds = data[i]

        # Pressure [GPa] grid
        if module == "aragog":
            xx_pres = ds["pres_b"]
        elif module == "spider":
            xx_pres = ds.get_dict_values(['data','pressure_b']) * 1e-9

        # Phase masks
        if module == "aragog":
            yy = np.array(ds["phi_b"])
            MASK_SO = yy < 0.05
            MASK_MI = (0.05 <= yy) & ( yy <= 0.95)
            MASK_ME = yy > 0.95
        elif module == "spider":
            MASK_MI = ds.get_mixed_phase_boolean_array( 'basic' )
            MASK_ME = ds.get_melt_phase_boolean_array(  'basic' )
            MASK_SO = ds.get_solid_phase_boolean_array( 'basic' )

        # Widen mixed phase region by 1 index so that lines are continuous on the plot
        MASK_MI[2:-2] = MASK_MI[2:-2] | MASK_MI[1:-3] | MASK_MI[3:-1]

        # Depth [km] grid
        if module == "aragog":
            xx_radius = ds["radius_b"][:]
            xx_depth = xx_radius[-1] - xx_radius
        elif module == "spider":
           xx_radius = ds.get_dict_values(['data','radius_b']) * 1e-3
           xx_depth = xx_radius[0] - xx_radius
        xx_depth = np.array(xx_depth, dtype=float) # make copy

        # Plot temperature
        if module == "aragog":
            yy = ds["temp_b"][:]
        elif module == "spider":
            yy = ds.get_dict_values(['data','temp_b'])
        yy = np.array(yy, dtype=float) / 1e3 # convert to kK

        axs[0].plot( yy[MASK_SO], xx_pres[MASK_SO], ls='solid',  c=color, lw=lw, label=label )
        axs[0].plot( yy[MASK_MI], xx_pres[MASK_MI], ls='dashed', c=color, lw=lw)
        axs[0].plot( yy[MASK_ME], xx_pres[MASK_ME], ls='dotted', c=color, lw=lw)

        # Plot melt fraction
        if module == "aragog":
            yy = ds["phi_b"][:]
        elif module == "spider":
            yy = ds.get_dict_values(['data','phi_b'])
        yy *= 100.0 # convert to percentage
        axs[1].plot( yy[MASK_SO], xx_pres[MASK_SO], ls='solid',  c=color, lw=lw)
        axs[1].plot( yy[MASK_MI], xx_pres[MASK_MI], ls='dashed',  c=color, lw=lw)
        axs[1].plot( yy[MASK_ME], xx_pres[MASK_ME], ls='dotted',  c=color, lw=lw)

        # Plot viscosity
        if module == "aragog":
            yy = 10**ds["log10visc_b"][:]
        elif module == "spider":
            visc_const = 1 # this is used for the arcsinh scaling
            yy = ds.get_dict_values(['data','visc_b'], MyFuncFormatter( visc_const ))
        axs[2].plot( yy[MASK_SO], xx_pres[MASK_SO], ls='solid',  c=color, lw=lw)
        axs[2].plot( yy[MASK_MI], xx_pres[MASK_MI], ls='dashed',  c=color, lw=lw)
        axs[2].plot( yy[MASK_ME], xx_pres[MASK_ME], ls='dotted',  c=color, lw=lw)
        visc_min = min(visc_min, np.amin(yy))
        visc_max = max(visc_max, np.amax(yy))

        # Plot convective flux
        if module == "aragog":
            yy = ds["Fconv_b"][:]
        elif module == "spider":
            yy =  ds.get_dict_values(['data','Jconv_b'])
        axs[3].plot( yy[MASK_SO], xx_pres[MASK_SO], ls='solid',   c=color, lw=lw)
        axs[3].plot( yy[MASK_MI], xx_pres[MASK_MI], ls='dashed',  c=color, lw=lw)
        axs[3].plot( yy[MASK_ME], xx_pres[MASK_ME], ls='dotted',  c=color, lw=lw)
        flux_min = min(flux_min, np.amin(yy))
        flux_max = max(flux_max, np.amax(yy))

    # Decorate figure
    title = '(a) Temperature' #'(a) Temperature, {}'.format(units)
    axs[0].set( title=title, xlabel=r'$T$ [kK]', ylabel=r'$P$ [GPa]')
    axs[0].set_ylim(top=np.amin(xx_pres), bottom=np.amax(xx_pres))
    axs[0].yaxis.set_minor_locator(MultipleLocator(10.0))
    axs[0].xaxis.set_minor_locator(MultipleLocator(0.25))
    axs[0].legend( fontsize=8, fancybox=True, framealpha=0.9, loc='lower left')

    title = '(b) Melt fraction'
    axs[1].set(title=title, xlabel=r'$\phi$ [%]')
    axs[1].set_xlim(left=-5, right=105)
    axs[1].xaxis.set_major_locator(MultipleLocator(25))
    axs[1].xaxis.set_minor_locator(MultipleLocator(5))

    axs[1].plot([-100,-200],[-100,-200], ls='solid',  c='k', lw=lw, label="Solid")
    axs[1].plot([-100,-200],[-100,-200], ls='dashed', c='k', lw=lw, label="Mush")
    axs[1].plot([-100,-200],[-100,-200], ls='dotted', c='k', lw=lw, label="Melt")
    axs[1].legend(fontsize=8, fancybox=True, framealpha=0.9, loc='lower left')

    title = '(c) Viscosity'
    axs[2].set( title=title, xlabel=r'$\eta$ [Pa s]')
    if visc_max > 100.0*visc_min:
        axs[2].set_xscale("log")

    title = '(d) Convective flux'
    axs[3].set( title=title, xlabel=r'$F_c$ [W m$^{-2}$]')
    if flux_max > 100.0*flux_min:
        axs[3].set_xscale("symlog", linthresh=1.0)

    # Pressure-depth conversion for y-axis
    ax3b = axs[3].twinx()
    ax3b.plot( yy, xx_depth, alpha=0.0)
    ax3b.set_ylim(top=np.amin(xx_depth), bottom=np.amax(xx_depth))
    ax3b.yaxis.set_minor_locator(MultipleLocator(100.0))
    ax3b.set_ylabel( '$d$ [km]')

    # Save figure
    fig.subplots_adjust(wspace=0.05)
    plt.close()
    plt.ioff()
    fpath = os.path.join(output_dir, "plot_interior.%s"%plot_format)
    fig.savefig(fpath, dpi=200, bbox_inches='tight')


def plot_interior_entry(handler: Proteus):
    module = handler.config.interior.module
    if module == "spider":
        extension = ".json"
    elif module == "aragog":
        extension = "_int.nc"
    plot_times,_ = sample_output(handler, extension=extension, tmin=1e3)
    print("Snapshots:", plot_times)

    data = read_interior_data(handler.directories["output"], module, plot_times)

    plot_interior(
        output_dir=handler.directories['output'],
        times=plot_times,
        data=data,
        module=module,
        plot_format=handler.config.params.out.plot_fmt,
    )


if __name__ == "__main__":
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_interior_entry(handler)



# Function and classes used to run CALLIOPE
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from proteus.config import Config

from calliope.constants import molar_mass, ocean_moles
from calliope.solve import (
    equilibrium_atmosphere,
    get_target_from_params,
    get_target_from_pressures,
)

from proteus.utils.constants import element_list, vol_list
from proteus.utils.helper import UpdateStatusfile

log = logging.getLogger("fwl."+__name__)

# Constants
mass_ocean = ocean_moles * molar_mass['H2']

def construct_options(dirs:dict, config:Config, hf_row:dict):
    """
    Construct CALLIOPE options dictionary
    """

    invalid = False  # Invalid options set by config

    solvevol_inp = {}

    # Planet properties
    solvevol_inp["M_mantle"]    = hf_row["M_mantle"]
    solvevol_inp["Phi_global"]  = hf_row["Phi_global"]
    solvevol_inp["gravity"]     = hf_row["gravity"]
    solvevol_inp["radius"]      = hf_row["R_int"]

    # Surface properties
    solvevol_inp["T_magma"]     =  hf_row["T_magma"]
    solvevol_inp['fO2_shift_IW'] = config.outgas.fO2_shift_IW

    # Calculate hydrogen inventory...

    #    absolute part (H_kg = H_oceans * number_ocean_moles * molar_mass['H2'])
    H_abs = float(config.delivery.elements.H_oceans) * mass_ocean

    #    relative part (H_kg = H_rel * 1e-6 * M_mantle)
    H_rel = config.delivery.elements.H_ppmw * 1e-6 * hf_row["M_mantle"]

    #    use whichever was set (one of these will be zero)
    if H_abs < 1.0:
        if H_rel < 1.0:
            log.error("Hydrogen inventory is unspecified")
            invalid = True
        else:
            H_kg = H_rel
    elif H_rel < 1.0:
        H_kg = H_abs
    else:
        log.error("Hydrogen inventory must be specified by H_oceans or H_ppmw, not both")
        invalid = True
        H_kg = -1 # dummy value

    # Calculate carbon inventory (we need CH_ratio for calliope)
    CH_ratio = float(config.delivery.elements.CH_ratio)
    C_ppmw   = float(config.delivery.elements.C_ppmw)
    if CH_ratio > 1e-10:
        # check that C_ppmw isn't also set
        if C_ppmw > 1e-10:
            log.error("Carbon inventory must be specified by CH_ratio or C_ppmw, not both")
            invalid = True
    else:
        if C_ppmw < 1e-10:
            log.error("Carbon inventory is unspecified")
            invalid = True
        else:
            # calculate C/H ratio for calliope from C_kg and H_kg
            CH_ratio = config.delivery.elements.C_ppmw * 1e-6 * hf_row["M_mantle"] / H_kg

    # Calculate nitrogen inventory (we need N_ppmw for calliope)
    NH_ratio = float(config.delivery.elements.NH_ratio)
    N_ppmw   = float(config.delivery.elements.N_ppmw)
    if NH_ratio > 1e-10:
        # check that N_ppmw isn't also set
        if N_ppmw > 1e-10:
            log.error("Nitrogen inventory must be specified by NH_ratio or N_ppmw, not both")
            invalid = True
        # calculate N_ppmw
        N_ppmw = 1e6 * NH_ratio * H_kg / hf_row["M_mantle"]
    else:
        if N_ppmw < 1e-10:
            log.error("Nitrogen inventory is unspecified")
            invalid = True


    # Calculate sulfur inventory (we need S_ppmw for calliope)
    SH_ratio = float(config.delivery.elements.SH_ratio)
    S_ppmw   = float(config.delivery.elements.S_ppmw)
    if SH_ratio > 1e-10:
        # check that S_ppmw isn't also set
        if S_ppmw> 1e-10:
            log.error("Sulfur inventory must be specified by SH_ratio or S_ppmw, not both")
            invalid = True
        # calculate S_ppmw
        S_ppmw = 1e6 * SH_ratio * H_kg / hf_row["M_mantle"]
    else:
        if S_ppmw < 1e-10:
            log.error("Sulfur inventory is unspecified")
            invalid = True

    # Volatile abundances are over-specified in the config file.
    # The code exits here, rather than above, in case there are multiple
    #   instances of volatiles being over-specified in the file.
    if invalid:
        log.error("  a) set X by metallicity, e.g. XH_ratio=1.2 and X_ppmw=0")
        log.error("  b) set X by concentration, e.g. XH_ratio=0 and X_ppmw=2.01")
        UpdateStatusfile(dirs, 20)
        exit(1)

    # Pass elemental inventory
    solvevol_inp['hydrogen_earth_oceans'] = H_kg / mass_ocean
    solvevol_inp['CH_ratio']    =           CH_ratio
    solvevol_inp['nitrogen_ppmw'] =         N_ppmw
    solvevol_inp['sulfur_ppmw'] =           S_ppmw

    # Volatile inventory
    for s in vol_list:
        solvevol_inp[f'{s}_initial_bar'] = config.delivery.volatiles.get_pressure(s)

        included = config.outgas.calliope.is_included(s)
        solvevol_inp[f'{s}_included'] = int(included)

        if (s in ("H2O","CO2","N2","S2")) and not included:
            UpdateStatusfile(dirs, 20)
            log.error(f"Missing required volatile {s}")
            exit(1)

    return solvevol_inp


def calc_target_masses(dirs:dict, config:Config, hf_row:dict):
    # make solvevol options
    solvevol_inp = construct_options(dirs, config, hf_row)

    # calculate target mass of atoms (except O, which is derived from fO2)
    if config.delivery.initial == 'elements':
        solvevol_target = get_target_from_params(solvevol_inp)
    else:
        solvevol_target = get_target_from_pressures(solvevol_inp)

    # store in hf_row as elements
    for e in solvevol_target.keys():
        if e == "O":
            continue
        hf_row[e + "_kg_total"] = solvevol_target[e]

def calc_surface_pressures(dirs:dict, config:Config, hf_row:dict):
    # make solvevol options
    solvevol_inp = construct_options(dirs, config, hf_row)

    # convert masses to dict for calliope
    solvevol_target = {}
    for e in element_list:
        if e == "O":
            continue

        # save to dict
        solvevol_target[e] = hf_row[e + "_kg_total"]

    # Do not allow low temperatures
    if solvevol_inp["T_magma"] < config.outgas.calliope.T_floor:
        solvevol_inp["T_magma"] = config.outgas.calliope.T_floor
        log.warning("Outgassing temperature clipped to %.1f K"%solvevol_inp["T_magma"])

    # get atmospheric compositison
    solvevol_result = equilibrium_atmosphere(solvevol_target, solvevol_inp)
    for k in solvevol_result.keys():
        if k in hf_row.keys():
            hf_row[k] = solvevol_result[k]



# Functions used to help run PROTEUS which are mostly submodule agnostic.

# Import utils-specific modules
from __future__ import annotations

import glob
import logging
import os
import subprocess
from datetime import datetime
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from proteus.atmos_clim.common import read_atmosphere_data
from proteus.interior.wrapper import read_interior_data
from proteus.plot.cpl_atmosphere import plot_atmosphere
from proteus.plot.cpl_elements import plot_elements
from proteus.plot.cpl_emission import plot_emission
from proteus.plot.cpl_escape import plot_escape
from proteus.plot.cpl_fluxes_atmosphere import plot_fluxes_atmosphere
from proteus.plot.cpl_fluxes_global import plot_fluxes_global
from proteus.plot.cpl_global import plot_global
from proteus.plot.cpl_interior import plot_interior
from proteus.plot.cpl_interior_cmesh import plot_interior_cmesh
from proteus.plot.cpl_observables import plot_observables
from proteus.plot.cpl_population import (
    plot_population_mass_radius,
    plot_population_time_density,
)
from proteus.plot.cpl_sflux import plot_sflux
from proteus.plot.cpl_sflux_cross import plot_sflux_cross
from proteus.plot.cpl_structure import plot_structure
from proteus.utils.constants import (
    element_list,
    gas_list,
)
from proteus.utils.helper import UpdateStatusfile, get_proteus_dir, safe_rm
from proteus.utils.plot import sample_times

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def _get_current_time():
    '''
    Get the current system time as a formatted string.
    '''
    return str(datetime.now().strftime('%Y-%m-%d %H:%M:%S %Z'))

def _get_git_revision(dir:str) -> str:
    '''
    Get git hash for repository in `dir`.
    '''
    # change dir
    cwd = os.getcwd()
    os.chdir(dir)

    # get hash (https://stackoverflow.com/a/21901260)
    hash = subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode('ascii').strip()

    # change dir back
    os.chdir(cwd)

    return hash

def _get_socrates_version():
    '''
    Get the installed SOCRATES version.
    '''
    verpath = os.path.join(os.environ.get("RAD_DIR"),"version")
    with open(verpath, 'r') as hdl:
        ver = hdl.read().replace("\n","")
    return str(ver)

def _get_spider_version():
    '''
    Get the installed SPIDER version.
    '''

    # This is the only one that we use, and it won't be updated in the future.
    return "0.2.0"

def _get_petsc_version():
    '''
    Get the installed PETSc version.
    '''

    # Like SPIDER, this is the only one that we use
    return "1.3.19.0"

def _get_agni_version(dirs:dict):
    '''
    Get the installed AGNI version
    '''
    from tomllib import load as tomlload
    with open(os.path.join(dirs["agni"],"Project.toml"),'rb') as hdl:
        agni_meta = tomlload(hdl)
    return agni_meta["version"]

def _get_julia_version():
    '''
    Get the installed Julia version
    '''
    return subprocess.check_output(["julia","--version"]).decode("utf-8").split()[-1]

def print_system_configuration(dirs:dict):
    '''
    Print the current system configuration.
    '''
    import platform
    import pwd
    import sys

    # Try to get the login name using os.getlogin()
    try:
        username = os.getlogin()
    except OSError:
        username = pwd.getpwuid(os.getuid()).pw_name

    log.info("Current time      " + _get_current_time())
    log.info("Python version    " + sys.version.split(" ")[0])
    log.info("System hostname   " + str(os.uname()[1]))
    log.info("System username   " + str(username))
    log.info("Platform type     " + str(platform.system()))
    log.info("FWL data path     " + dirs["fwl"])
    log.info(" ")

def print_module_configuration(dirs:dict, config:Config, config_path:str):
    '''
    Print the current module configuration, with versions.
    '''

    # PROTEUS
    from proteus import __version__ as proteus_version
    log.info("PROTEUS version   " + proteus_version)
    log.info("PROTEUS location  " + dirs["proteus"])
    log.info("PROTEUS git hash  " + _get_git_revision(dirs["proteus"]))
    log.info("Config file       " + str(config_path))
    log.info("Output path       " + dirs["output"])
    log.info(" ")

    # Interior module
    write = "Interior module   %s" % config.interior.module
    match config.interior.module:
        case 'spider':
            write += " version " + _get_spider_version()
        case 'aragog':
            from aragog import __version__ as aragog_version
            write += " version " + aragog_version
    log.info(write)
    if config.interior.module == "spider":
        log.info("  - PETSc         version " + _get_petsc_version())

    # Atmosphere module
    write = "Atmos_clim module %s" % config.atmos_clim.module
    match config.atmos_clim.module:
        case 'janus':
            from janus import __version__ as janus_version
            write += " version " + janus_version
        case 'agni':
            write += " version " + _get_agni_version(dirs)
    log.info(write)
    if config.atmos_clim.module in ['janus', 'agni']:
        log.info("  - SOCRATES      version %s at %s" %(_get_socrates_version(), dirs["rad"]))
        if config.atmos_clim.module == 'agni':
            log.info("  - Julia         version " + _get_julia_version())

    # Outgassing module
    write = "Outgas module     %s" % config.outgas.module
    if config.outgas.module == 'calliope':
        from calliope import __version__ as calliope_version
        write += " version " + calliope_version
    log.info(write)

    # Escape module
    log.info("Escape module     %s" % config.escape.module)

    # Star module
    write = "Star module       %s" % config.star.module
    if config.star.module == 'mors':
        from mors import __version__ as mors_version
        write += " version " + mors_version
    log.info(write)

    # Orbit module
    log.info("Orbit module      %s" % config.orbit.module)

    # Delivery module
    log.info("Delivery module   %s" % config.delivery.module)

    # End spacer
    log.info(" ")


def print_header():
    log.info(":::::::::::::::::::::::::::::::::::::::::::::::::::::::")
    log.info("                   PROTEUS framework                   ")
    log.info("                   by Forming Worlds                   ")
    log.info(":::::::::::::::::::::::::::::::::::::::::::::::::::::::")
    log.info(" ")

def PrintCurrentState(hf_row:dict):
    '''
    Print the current state of the model to the logger
    '''
    log.info("Runtime info...")
    log.info("    System time  :   %s  "         % _get_current_time())
    log.info("    Model time   :   %.2e   yr"    % float(hf_row["Time"]))
    log.info("    T_surf       :   %4.3f   K"    % float(hf_row["T_surf"]))
    log.info("    T_magma      :   %4.3f   K"    % float(hf_row["T_magma"]))
    log.info("    P_surf       :   %.2e   bar"   % float(hf_row["P_surf"]))
    log.info("    Phi_global   :   %.2e   "      % float(hf_row["Phi_global"]))
    log.info("    Instellation :   %.2e   W m-2" % float(hf_row["F_ins"]))
    log.info("    F_int        :   %.2e   W m-2" % float(hf_row["F_int"]))
    log.info("    F_atm        :   %.2e   W m-2" % float(hf_row["F_atm"]))
    log.info("    |F_net|      :   %.2e   W m-2" % abs(float(hf_row["F_net"])))

def CreateLockFile(output_dir:str):
    '''
    Create a lock file which, if removed, will signal for the simulation to stop.
    '''
    keepalive_file = os.path.join(output_dir,"keepalive")
    safe_rm(keepalive_file)
    with open(keepalive_file, 'w') as fp:
        fp.write("Removing this file will be interpreted by PROTEUS as a request to stop the simulation loop\n")
    return keepalive_file

def GetHelpfileKeys():
    '''
    Variables to be held in the helpfile.

    All dimensional quantites should be stored in SI units, except those noted below.
    * Pressure is in units of [bar].
    * Time is in units of [years].
    '''

    # Basic keys
    keys = [
            # Model tracking
            "Time", # [yr]

            # Orbital dynamics
            "separation", # [m]
            "period", # [s]

            # Dry interior radius (calculated) and mass (from config)
            "R_int", "M_int", # [m], [kg]

            # Total planet mass (mantle + core + volatiles)
            "M_tot", # [kg]

            # Temperatures
            "T_surf", "T_magma", "T_eqm", "T_skin", # all [K]

            # Energy fluxes, all [W m-2]
            "F_int", "F_atm", "F_net", "F_olr", "F_sct", "F_ins", "F_tidal", "F_radio",

            # Interior properties
            "gravity", "Phi_global", "RF_depth", # [m s-2] , [1] , [1]
            "M_core", "M_mantle", "M_planet",    # all [kg]
            "M_mantle_solid", "M_mantle_liquid", # all [kg]

            # Stellar
            "M_star", "R_star", "age_star", # [kg], [m], [yr]

            # Observational (from infinity)
            "p_obs",    # observered radius [bar]
            "z_obs",    # observed height relative to R_int [m]
            "rho_obs",  # observed bulk density [kg m-3]
            "transit_depth", "contrast_ratio", # [1], [1]
            "bond_albedo", # bond albedo [1]

            # Escape
            "esc_rate_total", "p_xuv", "z_xuv", "R_xuv", # [kg s-1], [bar], [m], [m]

            # Atmospheric composition
            "M_atm", "P_surf", "atm_kg_per_mol", # [kg], [bar], [kg mol-1]
            ]

    # gases
    for s in gas_list:
        keys.append(s+"_mol_atm")
        keys.append(s+"_mol_solid")
        keys.append(s+"_mol_liquid")
        keys.append(s+"_mol_total")
        keys.append(s+"_kg_atm")
        keys.append(s+"_kg_solid")
        keys.append(s+"_kg_liquid")
        keys.append(s+"_kg_total")
        keys.append(s+"_vmr")
        keys.append(s+"_bar")

    # element masses
    for e in element_list:
        keys.append(e+"_kg_atm")
        keys.append(e+"_kg_solid")
        keys.append(e+"_kg_liquid")
        keys.append(e+"_kg_total")

    # elemental ratios
    for e1 in element_list:
        for e2 in element_list:
            if e1==e2:
                continue

            # reversed ratio
            k = "%s/%s_atm"%(e2,e1)
            if k in keys:
                # skip this, since it's redundant to store (for example) the
                # ratio of H/C when we already have C/H.
                continue

            # intended ratio to be stored
            k = "%s/%s_atm"%(e1,e2)
            if k in keys:
                continue
            keys.append(k)

    return keys

def CreateHelpfileFromDict(d:dict):
    '''
    Create helpfile to hold output variables.
    '''
    log.debug("Creating new helpfile from dict")
    return pd.DataFrame([d], columns=GetHelpfileKeys(), dtype=float)

def ZeroHelpfileRow():
    '''
    Get a dictionary with same keys as helpfile but with values of zero
    '''
    out = {}
    for k in GetHelpfileKeys():
        out[k] = 0.0
    return out

def ExtendHelpfile(current_hf:pd.DataFrame, new_row:dict):
    '''
    Extend helpfile with new row of variables
    '''
    log.debug("Extending helpfile with new row")

    # validate keys
    missing_keys = set(GetHelpfileKeys()) - set(new_row.keys())
    if len(missing_keys)>0:
        raise Exception("There are mismatched keys in helpfile: %s"%missing_keys)

    # convert row to df
    new_row = pd.DataFrame([new_row], columns=GetHelpfileKeys(), dtype=float)

    # concatenate and return
    return pd.concat([current_hf, new_row], ignore_index=True)


def WriteHelpfileToCSV(output_dir:str, current_hf:pd.DataFrame):
    '''
    Write helpfile to a CSV file
    '''
    log.debug("Writing helpfile to CSV file")

    # check for invalid or missing keys
    difference = set(GetHelpfileKeys()) - set(current_hf.keys())
    if len(difference) > 0:
        raise Exception("There are mismatched keys in helpfile: "+str(difference))

    # remove old file
    fpath = os.path.join(output_dir , "runtime_helpfile.csv")
    if os.path.exists(fpath):
        os.remove(fpath)

    # write new file
    current_hf.to_csv(fpath, index=False, sep="\t", float_format="%.6e")
    return fpath

def ReadHelpfileFromCSV(output_dir:str):
    '''
    Read helpfile from disk CSV file to DataFrame
    '''
    fpath = os.path.join(output_dir , "runtime_helpfile.csv")
    if not os.path.exists(fpath):
        raise Exception("Cannot find helpfile at '%s'"%fpath)
    return pd.read_csv(fpath, sep=r"\s+")

def UpdatePlots( hf_all:pd.DataFrame, dirs:dict, config:Config, end=False, num_snapshots=7):
    """Update plots during runtime for analysis

    Calls various plotting functions which show information about the interior/atmosphere's energy and composition.

    Parameters
    ----------
        dirs : dict
            Directories dictionary
        config : Config
            PROTEUS options dictionary.
        end : bool
            Is this function being called at the end of the simulation?
    """

    # Directories
    output_dir = dirs["output"]
    fwl_dir    = dirs["fwl"]

    # Check model configuration
    dummy_atm = config.atmos_clim.module == 'dummy'
    dummy_int = config.interior.module == 'dummy'
    spider    = config.interior.module == 'spider'
    aragog    = config.interior.module == 'aragog'
    escape    = config.escape.module is not None

    # Get all output times
    output_times = []
    if spider:
        from proteus.interior.spider import get_all_output_times
        output_times = get_all_output_times( output_dir )
    if aragog:
        from proteus.interior.aragog import get_all_output_times
        output_times = get_all_output_times( output_dir )

    # Global properties for all timesteps
    plot_global(hf_all, output_dir, config)

    # Elemental mass inventory
    if escape:
        plot_elements(hf_all, output_dir, config.params.out.plot_fmt)
        plot_escape(hf_all, output_dir, escape_model=config.escape.module, plot_format=config.params.out.plot_fmt)

    # Which times do we have atmosphere data for?
    if not dummy_atm:
        ncs = glob.glob(os.path.join(output_dir, "data", "*_atm.nc"))
        nc_times = [int(f.split("/")[-1].split("_atm")[0]) for f in ncs]

        # Check intersection of atmosphere and interior data
        if dummy_int:
            output_times = nc_times
        else:
            output_times = sorted(list(set(output_times) & set(nc_times)))

    # Samples for plotting profiles
    if len(output_times) > 0:
        nsamp = 7
        tmin = 1.0
        if np.amax(output_times) > 1e3:
            tmin = 1e3
        plot_times, _ = sample_times(output_times, nsamp, tmin=tmin)
        log.debug("Snapshots to plot:" + str(plot_times))

    # Interior profiles
    if not dummy_int:
        int_data = read_interior_data(output_dir, config.interior.module, plot_times)
        plot_interior(output_dir, plot_times, int_data,
                              config.interior.module, config.params.out.plot_fmt)

    # Temperature profiles
    if not dummy_atm:
        atm_data = read_atmosphere_data(output_dir, plot_times)

        # Atmosphere only
        plot_atmosphere(output_dir, plot_times, atm_data, config.params.out.plot_fmt)

        # Atmosphere and interior, stacked
        if not dummy_int:
            plot_structure(hf_all, output_dir, plot_times, int_data, atm_data,
                            config.interior.module, config.params.out.plot_fmt)

        # Flux profiles
        if config.atmos_clim.module == 'janus':
            # only do this for JANUS, AGNI does it automatically
            plot_fluxes_atmosphere(output_dir, config.params.out.plot_fmt)

    # Only at the end of the simulation
    if end:
        plot_global(hf_all,         output_dir, config, logt=False)
        plot_fluxes_global(hf_all,  output_dir, config)
        plot_observables(hf_all,    output_dir, plot_format=config.params.out.plot_fmt)
        plot_population_mass_radius (hf_all, output_dir, fwl_dir, config.params.out.plot_fmt)
        plot_population_time_density(hf_all, output_dir, fwl_dir, config.params.out.plot_fmt)

        plt_modern = bool(config.star.module == "mors")
        if plt_modern:
            modern_age = config.star.mors.age_now * 1e9
        else:
            modern_age = -1
        plot_sflux(output_dir, plt_modern=plt_modern,
                            plot_format=config.params.out.plot_fmt)
        plot_sflux_cross(output_dir,modern_age=modern_age,
                            plot_format=config.params.out.plot_fmt)

        if not dummy_int:
            plot_interior_cmesh(output_dir, plot_times, int_data, config.interior.module,
                                plot_format=config.params.out.plot_fmt)

        if not dummy_atm:
            plot_emission(output_dir, plot_times, plot_format=config.params.out.plot_fmt)

    # Close all figures
    plt.close()


def get_proteus_directories(*, out_dir: str = 'proteus_out') -> dict[str, str]:
    """Create dict of proteus directories from root dir.

    Parameters
    ----------
    root_dir : str
        Proteus root directory
    out_dir : str, optional
        Name out output directory

    Returns
    -------
    dirs : dict[str, str]
        Proteus directories dict
    """
    root_dir = get_proteus_dir()

    return {
        "agni": os.path.join(root_dir, "AGNI"),
        "input": os.path.join(root_dir, "input"),
        "output": os.path.join(root_dir, "output", out_dir),
        "proteus":  root_dir,
        "spider": os.path.join(root_dir, "SPIDER"),
        "tools": os.path.join(root_dir, "tools"),
        "vulcan": os.path.join(root_dir, "VULCAN"),
        "utils": os.path.join(root_dir, "src", "proteus", "utils")
    }


def SetDirectories(config: Config) -> dict[str, str]:
    """Set directories dictionary

    Sets paths to the required directories, based on the configuration provided
    by the options dictionary.

    Parameters
    ----------
    config : Config
        PROTEUS options dictionary

    Returns
    ----------
    dirs : dict
        Dictionary of paths to important directories
    """
    dirs = get_proteus_directories(out_dir=config.params.out.path)

    # FWL data folder
    if os.environ.get('FWL_DATA') is None:
        UpdateStatusfile(dirs, 20)
        raise Exception("The FWL_DATA environment variable needs to be set up!"
                        "Did you source PROTEUS.env?")
    else:
        dirs["fwl"] = os.environ.get('FWL_DATA')

    # SOCRATES directory
    if config.atmos_clim.module in ('janus', 'agni'):
        # needed for atmosphere models 0 and 1

        if os.environ.get('RAD_DIR') is None:
            UpdateStatusfile(dirs, 20)
            raise Exception("The RAD_DIR environment variable has not been set")
        else:
            dirs["rad"] = os.environ.get('RAD_DIR')

    # Get abspaths
    for key in dirs.keys():
        dirs[key] = os.path.abspath(dirs[key])+"/"

    return dirs


# End of file



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from proteus.utils.constants import M_earth, element_list

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)

def plot_elements( hf_all:pd.DataFrame, output_dir:str, plot_format="pdf", t0=100.0):

    if np.amax(hf_all["Time"]) < 2:
        log.debug("Insufficient data to make plot_elements")
        return

    log.info("Plot elements")

    time = np.array(hf_all["Time"] )

    # make plot
    lw = 1.8
    scale = 1.1
    fig,ax = plt.subplots(1,1, figsize=(6*scale,4*scale))

    # By element
    total = np.zeros(len(time))
    for e in element_list:

        _lw = lw
        if e == 'H':
            _lw = lw * 1.8

        # Plot planetary inventory of this element
        y = np.array(hf_all[e+"_kg_total"])/M_earth
        total += y
        line = ax.plot(time, y, lw=_lw, ls='solid',  label=e)[0]

        # Plot atmospheric inventory of this element
        y = np.array(hf_all[e+"_kg_atm"])/M_earth
        ax.plot(time, y, lw=_lw, ls='dotted', color=line.get_color())

    # Planetary element sum inventory
    ax.plot(time, total, lw=lw, ls='solid',  label='Total',  c='k')

    # Atmosphere mass
    M_atm = np.array(hf_all["M_atm"])/M_earth
    ax.plot(time, M_atm, lw=lw, ls='dotted', label='Atmosphere', c='k')

    # decorate
    ax.set_ylabel(r"Mass [M$_\oplus$]")
    ax.set_yscale("log")
    ax.set_xlabel("Time [yr]")
    ax.set_xscale("log")
    ax.set_xlim(left=t0, right=np.amax(time)*1.1)
    ax.legend(loc='lower center', bbox_to_anchor=(0.5, 1.01), ncol=4)

    plt.close()
    plt.ioff()

    fpath = os.path.join(output_dir, "plot_elements.%s"%plot_format)
    fig.savefig(fpath, dpi=200, bbox_inches='tight')


def plot_elements_entry(handler: Proteus):
    # read helpfile
    hf_all = pd.read_csv(os.path.join(handler.directories['output'], "runtime_helpfile.csv"), sep=r"\s+")

    # make plot
    plot_elements(
        hf_all=hf_all,
        output_dir=handler.directories["output"],
        plot_format=handler.config.params.out.plot_fmt,
    )


if __name__ == "__main__":
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_elements_entry(handler)



# Aragog interior module
from __future__ import annotations

import glob
import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING

import netCDF4 as nc
import numpy as np
import pandas as pd
import platformdirs
from aragog import Output, Solver, aragog_file_logger
from aragog.parser import (
    Parameters,
    _BoundaryConditionsParameters,
    _EnergyParameters,
    _InitialConditionParameters,
    _MeshParameters,
    _PhaseMixedParameters,
    _PhaseParameters,
    _Radionuclide,
    _ScalingsParameters,
    _SolverParameters,
)

from proteus.interior.timestep import next_step
from proteus.utils.constants import R_earth, radnuc_data, secs_per_year

if TYPE_CHECKING:
    from proteus.config import Config

aragog_solver = None
log = logging.getLogger("fwl."+__name__)

FWL_DATA_DIR = Path(os.environ.get('FWL_DATA', platformdirs.user_data_dir('fwl_data')))

# Run the Aragog interior module
def RunAragog(config:Config, dirs:dict, IC_INTERIOR:int, hf_row:dict, hf_all:pd.DataFrame):

    global aragog_solver

    # Setup Aragog logger
    aragog_file_logger(console_level = logging.WARNING,
                       file_level = logging.INFO,
                       log_dir = dirs["output"])

    # Compute time step
    if IC_INTERIOR==1:
        dt = 0.0
        aragog_solver = None
    else:
        step_sf = 1.0 # dt scale factor
        dt = next_step(config, dirs, hf_row, hf_all, step_sf)

    # Setup Aragog parameters from options at first iteration
    if (aragog_solver is None):
        SetupAragogSolver(config, hf_row)
        # Update state from stored data if resuming a simulation
        if config.params.resume:
            UpdateAragogSolver(dt, hf_row, dirs["output"])
    # Update varying parameters in ongoing simulation
    else:
        UpdateAragogSolver(dt, hf_row)

    # Run Aragog solver
    aragog_solver.initialize()
    aragog_solver.solve()

    # Get Aragog output
    output = GetAragogOutput(hf_row)
    sim_time = aragog_solver.parameters.solver.end_time

    # Write output to a file
    WriteAragogOutput(dirs["output"],sim_time)

    return sim_time, output


def SetupAragogSolver(config:Config, hf_row:dict):

    global aragog_solver

    scalings = _ScalingsParameters(
            radius = R_earth, # scaling radius [m]
            temperature = 4000, # scaling temperature [K]
            density = 4000, # scaling density
            time = secs_per_year, # scaling time [sec]
            )

    solver = _SolverParameters(
            start_time = 0,
            end_time = 0,
            atol = config.interior.aragog.tolerance,
            rtol = config.interior.aragog.tolerance,
            )

    boundary_conditions = _BoundaryConditionsParameters(
            outer_boundary_condition = 4, # 4 = prescribed heat flux
            outer_boundary_value = hf_row["F_atm"], # first guess surface heat flux [W/m2]
            inner_boundary_condition = 1, # 1 = core cooling model, 3 = prescribed temperature
            inner_boundary_value = 4000, # core temperature [K], if inner_boundary_condition = 3
            emissivity = 1, # only used in gray body BC, outer_boundary_condition = 1
            equilibrium_temperature = hf_row["T_eqm"], # only used in gray body BC, outer_boundary_condition = 1
            core_density = config.struct.core_density, # used if inner_boundary_condition = 1
            core_heat_capacity = 880, # used if inner_boundary_condition = 1
            )

    mesh = _MeshParameters(
            outer_radius = hf_row["R_int"], # planet radius [m]
            inner_radius = config.struct.corefrac * hf_row["R_int"], # core radius [m]
            number_of_nodes = config.interior.aragog.num_levels, # basic nodes
            mixing_length_profile = "constant",
            surface_density = 4090, # AdamsWilliamsonEOS parameter [kg/m3]
            gravitational_acceleration = hf_row["gravity"], # [m/s-2]
            adiabatic_bulk_modulus = 260E9, # AdamsWilliamsonEOS parameter [Pa]
            )

    # Get tidal heat production [W/kg]
    tidal_value = 0.0
    if config.interior.tidal_heat and config.orbit.dummy:
        tidal_value = config.orbit.dummy.H_tide

    energy = _EnergyParameters(
            conduction = True,
            convection = True,
            gravitational_separation = False,
            mixing = False,
            radionuclides = config.interior.radiogenic_heat,
            tidal = config.interior.tidal_heat,
            tidal_value=tidal_value
            )

    initial_condition = _InitialConditionParameters(
            surface_temperature = config.interior.aragog.ini_tmagma, # initial top temperature (K)
            basal_temperature = config.interior.aragog.ini_tmagma, # initial bottom temperature (K)
            )

    # Get look up data directory, will be configurable in the future
    LOOK_UP_DIR = FWL_DATA_DIR / "interior_lookup_tables/1TPa-dK09-elec-free/MgSiO3_Wolf_Bower_2018/"

    phase_liquid = _PhaseParameters(
            density = LOOK_UP_DIR / "density_melt.dat",
            viscosity = 1E2,
            heat_capacity = LOOK_UP_DIR / "heat_capacity_melt.dat",
            melt_fraction = 1,
            thermal_conductivity = 4,
            #thermal_expansivity = 1.0E-5,
            thermal_expansivity = LOOK_UP_DIR / "thermal_exp_melt.dat",
            )

    phase_solid = _PhaseParameters(
            density = LOOK_UP_DIR / "density_solid.dat",
            viscosity = 1E21,
            heat_capacity = LOOK_UP_DIR / "heat_capacity_solid.dat",
            melt_fraction = 0,
            thermal_conductivity = 4,
            #thermal_expansivity = 1.0E-5,
            thermal_expansivity = LOOK_UP_DIR / "thermal_exp_solid.dat",
            )

    phase_mixed = _PhaseMixedParameters(
            latent_heat_of_fusion = 4e6,
            rheological_transition_melt_fraction = 0.4,
            rheological_transition_width = 0.15,
            solidus = LOOK_UP_DIR / "solidus.dat",
            liquidus = LOOK_UP_DIR / "liquidus.dat",
            phase = "mixed",
            phase_transition_width = 0.1,
            grain_size = config.interior.grain_size,
            )

    radionuclides = []
    if config.interior.radiogenic_heat:
        # offset by age_ini, which converts model simulation time to the actual age
        radio_t0 = config.delivery.radio_tref - config.star.age_ini
        radio_t0 *= 1e9 # Convert Gyr to yr

        def _append_radnuc(_iso, _cnc):
            radionuclides.append(_Radionuclide(
                                    name = _iso,
                                    t0_years = radio_t0,
                                    abundance = radnuc_data[_iso]["abundance"],
                                    concentration = _cnc,
                                    heat_production = radnuc_data[_iso]["heatprod"],
                                    half_life_years = radnuc_data[_iso]["halflife"],
                                ))

        if config.delivery.radio_K > 0.0:
            _append_radnuc("k40", config.delivery.radio_K)

        if config.delivery.radio_Th > 0.0:
            _append_radnuc("th232", config.delivery.radio_Th)

        if config.delivery.radio_U > 0.0:
            _append_radnuc("u235", config.delivery.radio_U)
            _append_radnuc("u238", config.delivery.radio_U)

    param = Parameters(
            boundary_conditions = boundary_conditions,
            energy = energy,
            initial_condition = initial_condition,
            mesh = mesh,
            phase_solid = phase_solid,
            phase_liquid = phase_liquid,
            phase_mixed = phase_mixed,
            radionuclides = radionuclides,
            scalings = scalings,
            solver = solver,
            )

    aragog_solver = Solver(param)

def UpdateAragogSolver(dt:float, hf_row:dict, output_dir:str = None):

    # Set solver time
    # hf_row["Time"] is in yr so do not need to scale as long as scaling time is secs_per_year
    aragog_solver.parameters.solver.start_time = hf_row["Time"]
    aragog_solver.parameters.solver.end_time = hf_row["Time"] + dt

    # Get temperature field from previous run
    if output_dir is not None: # read it from output directory
        Tfield = read_last_Tfield(output_dir, hf_row["Time"])
    else: # get it from solver
        Tfield = aragog_solver.temperature_staggered[:, -1]

    # Update initial condition
    Tfield = Tfield / aragog_solver.parameters.scalings.temperature
    aragog_solver.parameters.initial_condition.from_field = True
    aragog_solver.parameters.initial_condition.init_temperature = Tfield

    # Update boundary conditions
    aragog_solver.parameters.boundary_conditions.outer_boundary_value = hf_row["F_atm"]

    # TODO: Update tidal heating here when it is calculated self-consistently from
    #       the interior properties (e.g. viscosity) and evolved over time.

    return

def WriteAragogOutput(output_dir:str, time:float):

    aragog_output: Output = Output(aragog_solver)

    fpath = os.path.join(output_dir,"data","%d_int.nc"%time)
    aragog_output.write_at_time(fpath,-1)

def GetAragogOutput(hf_row:dict):

    aragog_output: Output = Output(aragog_solver)
    output = {}

    output["M_mantle"] = aragog_output.mantle_mass
    output["T_magma"] = aragog_output.solution_top_temperature
    output["Phi_global"] = float(aragog_output.melt_fraction_global)
    output["RF_depth"] = float(aragog_output.rheological_front)
    output["F_int"] = aragog_output.convective_heat_flux_basic[-1,-1] # Need to be revised for consistency

    output["M_mantle_liquid"] = output["M_mantle"] * output["Phi_global"]
    output["M_mantle_solid"] = output["M_mantle"] - output["M_mantle_liquid"]

    # Calculate surface area
    radii = aragog_output.radii_km_basic * 1e3 # [m]
    area  = 4 * np.pi * radii[-1]**2 # [m^2]

    # Mass at each mesh layer
    mass_s = aragog_output.mass_staggered[:,-1] # [kg]

    # Radiogenic heating
    Hradio_s = aragog_output.heating_radio[:,-1]  # [W kg-1]
    output["F_radio"] = np.dot(Hradio_s, mass_s) / area  # [W m-2]

    # Tidal heating
    Htidal_s = aragog_output.heating_tidal[:,-1]  # [W kg-1]
    output["F_tidal"] = np.dot(Htidal_s, mass_s)/area

    return output

def read_last_Tfield(output_dir:str, time:float):

    # Read Aragog output at last run
    fpath = os.path.join(output_dir,"data","%d_int.nc"%time)
    out = read_ncdf(fpath)

    # Get temperature field at basic nodes and interpolate to staggered nodes
    T_basic = out["temp_b"]
    N_basic = len(T_basic)
    T_staggered = (T_basic[0:N_basic-1] + T_basic[1:N_basic] ) / 2.0

    return T_staggered

def get_all_output_times(output_dir:str):
    files = glob.glob(output_dir+"/data/*_int.nc")
    years = [int(f.split("/")[-1].split("_int")[0]) for f in files]
    mask = np.argsort(years)

    return [years[i] for i in mask]

def read_ncdf(fpath:str):
    out = {}
    ds = nc.Dataset(fpath)

    for key in ds.variables.keys():
        out[key] = ds.variables[key][:]

    ds.close()
    return out

def read_ncdfs(output_dir:str, times:list):
    return [read_ncdf(os.path.join(output_dir, "data", "%d_int.nc"%t)) for t in times]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)

def plot_heatingrates(output_dir: str, atm, comp: list=[]):
    """Plot temperature structure and corresponding heating rates versus pressure

    Reads the temperature for each pressure level in atm, and plots it alongside
    the heating rates on the same pressure y-axis. Can also plot the temperature
    structure from another atmos object for comparison.

    Parameters
    ----------
        output_dir : str
            Output directory that was specified in the PROTEUS cfg file
        atm : atmos
            Atmosphere object containing P, T, heatingrates

        comp : np.ndarray shape (2,n)
            Optional PT profile for comparison (p=comp[0] and T=comp[1])
    """

    log.info("Plotting heating rates at each level")

    mpl.use('Agg')

    fig,ax = plt.subplots(1,1)

    ax.set_yscale("log")
    ax.invert_yaxis()
    ax.set_ylabel("Pressure [Bar]")

    ax2 = ax.twiny()
    c = 'tab:red'
    ax2.set_xlabel("Temperature [K]",color=c)
    ax2.tick_params(axis='x', colors=c)
    ax2.plot(atm.tmp,  atm.p  * 1.e-5, color=c, label="$T(p)$ centre", zorder=4)
    ax2.plot(atm.tmpl, atm.pl * 1.e-5, color=c, label="$T(p)$ edge",   zorder=3, alpha=0.5)

    if (len(comp) == 2):
        c = 'tab:purple'
        ax2.plot(comp[1],comp[0] * 1.e-5, color=c, label="$T(p)$ comparison",zorder=1)
        ax2.legend()

    c = 'black'
    ax.axvline(x=0,color=c,lw=0.4)
    ax.plot(atm.net_heating,atm.p * 1.e-5,color=c,zorder=2)
    ax.set_xlabel("Heating rate [K/day]")
    ax.set_xscale("symlog")
    ax.tick_params(axis='x', colors=c)

    plt.close()
    plt.ioff()
    fig.savefig(output_dir+"/plot_heatingrates.pdf")


def plot_heatingrates_entry(handler: Proteus):
    raise NotImplementedError(
        "`plot_heatingrates()` does not support command-line execution."
    )


if __name__ == '__main__':
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_heatingrates_entry(handler)



from __future__ import annotations

import importlib.metadata
import os
from functools import partial
from importlib.metadata import PackageNotFoundError
from typing import Callable

import click
import requests
from attr import dataclass

from proteus.utils.coupler import _get_agni_version, get_proteus_directories

DIRS = get_proteus_directories()

HEADER_STYLE = {'fg': 'yellow', 'underline': True, 'bold': True}
OK_STYLE = {'fg': 'green'}
ERROR_STYLE = {'fg': 'red'}
DEFAULT_STYLE = {'fg': 'yellow'}


@dataclass
class BasePackage:
    name: str

    def current_version(self) -> str: ...

    def latest_version(self) -> str: ...

    def get_status_message(self) -> str:
        try:
            current_version = self.current_version()
            latest_version = self.latest_version()
        except BaseException as exc:
            message = click.style(str(exc), **ERROR_STYLE)
        else:
            if current_version != latest_version:
                message = click.style(
                    f'Update available {current_version} -> {latest_version}', fg='yellow'
                )
            else:
                message = click.style('ok', **OK_STYLE)

        name = click.style(self.name, **OK_STYLE)
        return f'{name}: {message}'


class PythonPackage(BasePackage):
    def current_version(self) -> str:
        return importlib.metadata.version(self.name)

    def latest_version(self) -> str:
        response = requests.get(f'https://pypi.org/pypi/{self.name}/json')
        if not response.ok:
            response.raise_for_status()

        return response.json()['info']['version']


@dataclass
class GitPackage(BasePackage):
    owner: str
    version_getter: Callable

    def current_version(self) -> str:
        try:
            return self.version_getter()
        except FileNotFoundError as exc:
            raise PackageNotFoundError(f'{self.name} is not installed.') from exc

    def latest_version(self) -> str:
        response = requests.get(
            f'https://api.github.com/repos/{self.owner}/{self.name}/releases/latest'
        )
        return response.json()['name']


PACKAGES = (
    PythonPackage(name='aragog'),
    PythonPackage(name='fwl-calliope'),
    PythonPackage(name='fwl-janus'),
    PythonPackage(name='fwl-proteus'),
    PythonPackage(name='fwl-mors'),
    PythonPackage(name='fwl-zephyrus'),
    GitPackage(name='AGNI', owner='nichollsh', version_getter=partial(_get_agni_version, DIRS)),
)


def get_env_var_status_message(var: str) -> str:
    if os.environ.get(var):
        message = click.style('ok', **OK_STYLE)
    else:
        message = click.style('Variable not set.', **ERROR_STYLE)

    name = click.style(var, **OK_STYLE)
    return f'{name}: {message}'


VARIABLES = (
    'FWL_DATA',
    'RAD_DIR',
)


def doctor_entry():
    click.secho('Packages', **HEADER_STYLE)
    for package in PACKAGES:
        message = package.get_status_message()
        click.echo(message)

    click.secho('\nEnvironment variables', **HEADER_STYLE)
    for var in VARIABLES:
        message = get_env_var_status_message(var)
        click.echo(message)



from __future__ import annotations

import logging

import tomlkit
from attrs import asdict, define, field, validators

from ._atmos_clim import AtmosClim
from ._converters import dict_replace_none
from ._delivery import Delivery
from ._escape import Escape
from ._interior import Interior
from ._orbit import Orbit
from ._outgas import Outgas
from ._params import Params
from ._star import Star
from ._struct import Struct

log = logging.getLogger('fwl.' + __name__)

def spada_zephyrus(instance, attribute, value):
    # using zephyrus
    #     zephyrus requires MORS + Spada
    if (instance.escape.module == 'zephyrus') and \
        not ( (instance.star.module == 'mors') and (instance.star.mors.tracks == 'spada')):
        raise ValueError('ZEPHYRUS must be used with MORS and the Spada evolution tracks')

@define
class Config:
    """Root config parameters.

    Attributes
    ----------
    version: str
        Version of the configuration file.
    author: str
        Authors of the configuration file.
    params: Params
        Parameters for code execution, output files, time-stepping, convergence.
    star: Star
        Stellar parameters, model selection.
    orbit: Orbit
        Orbital and star-system parameters.
    struct: Struct
        Planetary structure calculation (mass, radius).
    atmos_clim: AtmosClim
        Planetary atmosphere parameters, model selection.
    escape: Escape
        Atmospheric escape parameters, model selection.
    interior: Interior
        Magma ocean / mantle model parameters, model selection.
    outgas: Outgas
        Outgassing parameters (fO2, etc) and included volatiles.
    delivery: Delivery
        Initial volatile inventory, and delivery model selection.
    """

    version: str = field(validator=validators.in_(('2.0',)))
    author: str

    params: Params
    star: Star
    orbit: Orbit
    struct: Struct
    atmos_clim: AtmosClim
    escape: Escape = field(validator=(spada_zephyrus,))
    interior: Interior
    outgas: Outgas
    delivery: Delivery

    def write(self, out:str):
        """
        Write configuration to a new TOML file.
        """

        # Convert to dictionary
        cfg = dict(asdict(self))

        # Replace None with "none"
        cfg = dict_replace_none(cfg)

        # Write to TOML file
        with open(out,'w') as hdl:
            tomlkit.dump(cfg, hdl)



# Functions used to handle atmosphere thermodynamics (running JANUS, etc.)
from __future__ import annotations

import glob
import logging
import os
import shutil
from typing import TYPE_CHECKING

import janus.set_socrates_env  # noqa
import numpy as np
import pandas as pd

from proteus.atmos_clim.common import get_height_from_pressure
from proteus.utils.constants import vap_list, vol_list
from proteus.utils.helper import UpdateStatusfile, create_tmp_folder

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def InitStellarSpectrum(dirs:dict, wl:list, fl:list, spectral_file_nostar):

    from janus.utils import InsertStellarSpectrum, PrepareStellarSpectrum

    log.debug("Prepare spectral file for JANUS")

    # Generate a new SOCRATES spectral file containing this new spectrum
    star_spec_src = dirs["output"]+"socrates_star.txt"

    # Spectral file stuff
    PrepareStellarSpectrum(wl,fl,star_spec_src)

    log.debug("Insert stellar spectrum into spectral file")
    InsertStellarSpectrum(spectral_file_nostar,
                          star_spec_src,
                          dirs["output"]
                          )
    os.remove(star_spec_src)

    return

def InitAtm(dirs:dict, config:Config):

    from janus.utils import ReadBandEdges, atmos

    log.debug("Create new JANUS atmosphere object")

    vol_dict = {}
    for vol in vol_list:
        vol_dict[vol] = 1.0/len(vol_list)

    # Spectral bands
    band_edges = ReadBandEdges(dirs["output"]+"star.sf")

    # Make object
    # The var flag indicates variable parameters
    # to be set at each PROTEUS iteration
    # through the routine UpdateStateAtm
    atm = atmos(0.0, #var
                1e5, #var
                config.atmos_clim.janus.p_top*1e5,
                6.371e6, #var
                5.972e24, #var
                band_edges,
                vol_mixing = vol_dict, #var
                req_levels = config.atmos_clim.janus.num_levels,
                water_lookup = False,
                alpha_cloud=config.atmos_clim.cloud_alpha,
                trppT = config.atmos_clim.tmp_minimum,
                minT = config.atmos_clim.tmp_minimum,
                maxT = config.atmos_clim.tmp_maximum,
                do_cloud = config.atmos_clim.cloud_enabled,
                re = 1.0e-5, # Effective radius of the droplets [m] (drizzle forms above 20 microns)
                lwm = 0.8, # Liquid water mass fraction [kg/kg]
                clfr = 0.8, # Water cloud fraction
                albedo_s = config.atmos_clim.surf_greyalbedo,
                albedo_pl = config.atmos_clim.albedo_pl,
                zenith_angle = config.orbit.zenith_angle,
                )

    atm.inst_sf = config.orbit.s0_factor
    atm.skin_d = config.atmos_clim.surface_d
    atm.skin_k = config.atmos_clim.surface_k

    match config.atmos_clim.janus.overlap_method:
        case "ro":
            atm.overlap_type = 2
        case "ee":
            atm.overlap_type = 4
        case "rorr":
            atm.overlap_type = 8
        case _:
            raise ValueError("Invalid overlap method selected for SOCRATES/JANUS!")

    return atm

def UpdateStateAtm(atm, hf_row:dict, tropopause):
    """UpdateStateAtm

    Update the atm object state with current iteration variables

    Parameters
    ----------
        atm : atmos
            Atmosphere object
        hf_row : dict
            Dictionary containing simulation variables for current iteration
        tropopause
            Tropopause type (None, "skin", "dynamic")
    """

    atm.setSurfaceTemperature(hf_row["T_surf"])
    atm.setSurfacePressure(hf_row["P_surf"]*1e5)
    atm.setPlanetProperties(hf_row["R_int"], hf_row["M_int"])

    # Warn about rock vapours
    has_vapours = False
    for gas in vap_list:
        has_vapours = has_vapours or (hf_row[gas+"_vmr"] > 1e-5)
    if has_vapours:
        log.warning("Atmosphere contains rock vapours which will be neglected by JANUS")

    # Store volatiles only
    vol_mixing = {}
    for vol in vol_list:
        vol_mixing[vol] = hf_row[vol+"_vmr"]
    atm.setVolatiles(vol_mixing)

    atm.instellation = hf_row["F_ins"]
    atm.tmp_magma = hf_row["T_magma"]
    if tropopause == "skin":
        atm.trppT = hf_row["T_skin"]
    else:
        atm.trppT = 0.5
    log.debug("Setting stratosphere to %.2f K"%atm.trppT)

    return

def RunJANUS(atm, dirs:dict, config:Config, hf_row:dict, hf_all:pd.DataFrame,
             write_in_tmp_dir=True, search_method=0, rtol=1.0e-4):
    """Run JANUS.

    Calculates the temperature structure of the atmosphere and the fluxes, etc.
    Stores the new flux boundary condition to be provided to SPIDER. Limits flux
    change if required.

    Parameters
    ----------
        atm : atmos
            Atmosphere object
        dirs : dict
            Dictionary containing paths to directories
        config : Config
            Configuration options and other variables
        hf_row : dict
            Dictionary containing simulation variables for current iteration
        hf_all : pd.DataFrame
            Dataframe containing simulation variables (now and historic)
        write_in_tmp_dir : bool
            Write temporary files in a local folder within /tmp, rather than in the output folder
        search_method : int
            Root finding method used by JANUS
        rtol : float
            Relative tolerance on solution for root finding method
    Returns
    ----------
        atm : atmos
            Updated atmos object
        output : dict
            Output variables, as a dict

    """

    from janus.utils.observed_rho import calc_observed_rho

    # Runtime info
    log.info("Running JANUS...")
    time = hf_row["Time"]


    #Update atmosphere with current variables
    UpdateStateAtm(atm, hf_row, config.atmos_clim.janus.tropopause)

    # Change dir
    cwd = os.getcwd()
    tmp_dir = dirs["output"]
    if write_in_tmp_dir:
        tmp_dir = create_tmp_folder()
    log.debug("Will run socrates inside '%s'"%tmp_dir)
    os.chdir(tmp_dir)

    # Prepare to calculate temperature structure w/ General Adiabat
    trppD = config.atmos_clim.janus.tropopause == 'dynamic'
    rscatter = config.atmos_clim.rayleigh

    # Run JANUS
    if config.atmos_clim.surf_state == 'fixed':  # fixed T_Surf
        from janus.modules import MCPA
        atm = MCPA(dirs, atm, False, trppD, rscatter)

    elif config.atmos_clim.surf_state == 'skin': # conductive lid
        from janus.modules import MCPA_CBL

        T_surf_max = -1
        T_surf_old = -1
        atol       = 1.0e-5

        # Done with initial loops
        if time > 0:

            # Get previous temperature as initial guess
            T_surf_old = hf_all.iloc[-1]["T_surf"]

            # Prevent heating of the interior
            if config.atmos_clim.prevent_warming:
                T_surf_max = T_surf_old

            # calculate tolerance
            tol = rtol * abs(hf_all.iloc[-1]["F_atm"]) + atol
        else:
            tol = 0.1

        # run JANUS
        atm = MCPA_CBL(dirs, atm, trppD, rscatter, method=search_method, atol=tol,
                        atm_bc=int(config.atmos_clim.janus.F_atm_bc), T_surf_guess=float(T_surf_old)-1.0, T_surf_max=float(T_surf_max))

    else:
        UpdateStatusfile(dirs, 20)
        raise Exception("Invalid surface state chosen for JANUS")

    # Clean up run directory
    for file in glob.glob(tmp_dir+"/current??.????"):
        os.remove(file)
    for file in glob.glob(tmp_dir+"/profile.*"):
        os.remove(file)
    os.chdir(cwd)
    if write_in_tmp_dir:
        shutil.rmtree(tmp_dir,ignore_errors=True)

    any_cloud = np.any(np.array(atm.clfr) > 1.0e-20)
    log.info("Water clouds have formed = %s"%(str(any_cloud)))
    log.info("SOCRATES fluxes (net@surf, net@TOA, OLR): %.5e, %.5e, %.5e W m-2" %
             (atm.net_flux[-1], atm.net_flux[0] , atm.LW_flux_up[0]))

    # Save atm data to disk
    nc_fpath = dirs["output"]+"/data/"+str(int(time))+"_atm.nc"
    atm.write_ncdf(nc_fpath)

    # Check for NaNs
    if not np.isfinite(atm.net_flux).all():
        UpdateStatusfile(dirs, 22)
        raise Exception("JANUS output array contains NaN or Inf values")

    # Store new flux
    if (config.atmos_clim.janus.F_atm_bc == 0):
        F_atm_new = atm.net_flux[0]
    else:
        F_atm_new = atm.net_flux[-1]

    # Require that the net flux must be upward
    F_atm_lim = F_atm_new
    if config.atmos_clim.prevent_warming:
        F_atm_lim = max( 1.0e-8 , F_atm_new )

    # Print if a limit was applied
    if not np.isclose(F_atm_lim , F_atm_new ):
        log.warning("Change in F_atm [W m-2] limited in this step!")
        log.warning("    %g  ->  %g" % (F_atm_new , F_atm_lim))

    # observables
    p_obs = float(config.atmos_clim.janus.p_obs)*1e5 # converted to Pa
    z_obs = 0.0
    rho_obs = -1.0
    if atm.height_error:
        log.error("Hydrostatic integration failed in JANUS!")
    else:
        # find observed level [m] at p ~ p_obs
        _, z_obs = get_height_from_pressure(atm.p, atm.z, p_obs)

        # calc observed density [kg m-3]
        rho_obs = calc_observed_rho(atm)

    # XUV height in atm
    if config.escape.module == 'zephyrus':
        # escape level set by zephyrus config
        p_xuv = config.escape.zephyrus.Pxuv # [bar]
    else:
        # escape level set to surface
        p_xuv = hf_row["P_surf"] # [bar]
    p_xuv, z_xuv = get_height_from_pressure(atm.p, atm.z, p_xuv*1e5) # [Pa], [m]

    # final things to store
    output={}
    output["T_surf"] = atm.ts            # Surface temperature [K]
    output["F_atm"]  = F_atm_lim         # Net flux at TOA
    output["F_olr"]  = atm.LW_flux_up[0] # OLR
    output["F_sct"]  = atm.SW_flux_up[0] # Scattered SW flux
    output["albedo"] = atm.SW_flux_up[0] / atm.SW_flux_down[0]
    output["p_obs"]  = p_obs/1e5        # observed level [bar]
    output["z_obs"]  = z_obs            # observed level [m]
    output["rho_obs"]= rho_obs          # observed density [kg m-3]
    output["p_xuv"]  = p_xuv/1e5         # Closest pressure from Pxuv    [bar]
    output["z_xuv"]  = z_xuv             # Height at Pxuv                [m]

    return output
